@@ -2,20 +2,27 @@ from __future__ import annotations
 
 import asyncio
 import inspect
-from typing import Any, AsyncIterator, Callable, Dict, List, Optional, TYPE_CHECKING, Union
+from typing import Any, AsyncIterator, Callable, Dict, Generator, List, Optional, TYPE_CHECKING, Union
 
 from aiocache import cached
 from aiocache.factory import Cache
 
 from mi import Instance, InstanceMeta, User
+from mi.actions import DriveActions, NoteActions, UserActions
+from mi.api.emoji import EmojiManager
+from mi.api.reaction import ReactionManager
 from mi.chat import Chat
-from mi.drive import Drive
+from mi.drive import File
 from mi.emoji import Emoji
-from mi.exception import ContentRequired, InvalidParameters, NotExistRequiredParameters
+from mi.exception import InvalidParameters, NotExistRequiredParameters
 from mi.http import Route
 from mi.iterators import InstanceIterator
-from mi.note import Note, Poll, Reaction
-from mi.user import Follower, Following
+from mi.models.chat import RawChat
+from mi.models.drive import RawFile
+from mi.models.note import RawNote
+from mi.models.user import RawUser
+from mi.note import Note, Reaction
+from mi.user import FollowRequest, Followee
 from mi.utils import check_multi_arg, get_cache_key, get_module_logger, key_builder, remove_dict_empty, str_lower, \
     upper_to_lower
 
@@ -24,167 +31,27 @@ if TYPE_CHECKING:
     from mi.types import (Note as NotePayload, Chat as ChatPayload)
 
 
-class NoteActions:
-    def __init__(self, http: HTTPClient, loop: asyncio.AbstractEventLoop):
-        self.http = http
-        self.loop = loop
+class ClientActions:
+    def __init__(self, state: 'ConnectionState', http: HTTPClient, loop: asyncio.AbstractEventLoop):
+        self.__state = state
+        self.__http = http
+        self.__loop = loop
+        self.note: NoteActions = NoteActions(state, http, loop)
+        self.user: UserActions = UserActions(state, http, loop)
+        self.drive: DriveActions = DriveActions(state, http, loop)
+        self.emoji: EmojiManager = EmojiManager(state, http, loop)
+        self.reaction: ReactionManager = ReactionManager(state, http, loop)
 
-    async def favorite(self, note_id: str) -> bool:
-        data = {'noteId': note_id}
-        return bool(await self.http.request(Route('POST', '/api/notes/favorites/create'), json=data, auth=True))
+    def get_user_instance(self, user_id: Optional[str]) -> UserActions:
+        return UserActions(self.__state, self.__http, self.__loop, user_id=user_id)
 
-    async def remove_favorite(self, note_id: str) -> bool:
-        data = {'noteId': note_id}
-        return bool(await self.http.request(Route('POST', '/api/notes/favorites/delete'), json=data, auth=True))
-
-    async def add_note_to_clips(self, clip_id: str, note_id: str) -> bool:
-        data = {'noteId': note_id, 'clipId': clip_id}
-        return bool(await self.http.request(Route('POST', '/api/clips/add-note'), json=data, auth=True))
-
-    async def add_reaction_to_note(self, note_id: str, reaction: str) -> bool:
-        data = {'noteId': note_id, 'reaction': reaction}
-        return bool(await self.http.request(Route('POST', '/api/reactions/create', json=data, auth=True)))
-
-    async def post_note(self,
-                        content: Optional[str] = None,
-                        visibility: str = "public",
-                        visible_user_ids: Optional[List[str]] = None,
-                        cw: Optional[str] = None,
-                        local_only: bool = False,
-                        no_extract_mentions: bool = False,
-                        no_extract_hashtags: bool = False,
-                        no_extract_emojis: bool = False,
-                        reply_id: Optional[str] = None,
-                        renote_id: Optional[str] = None,
-                        channel_id: Optional[str] = None,
-                        file_ids=None,
-                        poll: Optional[Poll] = None
-                        ):
-        """
-
-        ノートを投稿します。
-
-        Parameters
-        ----------
-        content : str
-            投稿する内容
-        visibility : str, optional
-            公開範囲, by default "public"
-        visible_user_ids : Optional[List[str]], optional
-            公開するユーザー, by default None
-        cw : Optional[str], optional
-            閲覧注意の文字, by default None
-        local_only : bool, optional
-            ローカルにのみ表示するか, by default False
-        no_extract_mentions : bool, optional
-            メンションを展開するか, by default False
-        no_extract_hashtags : bool, optional
-            ハッシュタグを展開するか, by default False
-        no_extract_emojis : bool, optional
-            絵文字を展開するか, by default False
-        reply_id : Optional[str], optional
-            リプライ先のid, by default None
-        renote_id : Optional[str], optional
-            リノート先のid, by default None
-        channel_id : Optional[str], optional
-            チャンネルid, by default None
-        file_ids : [type], optional
-            添付するファイルのid, by default None
-        poll : Optional[Poll], optional
-            アンケート, by default None
-
-        Returns
-        -------
-        [type]
-            [description]
-
-        Raises
-        ------
-        ContentRequired
-            [description]
-        """
-        if file_ids is None:
-            file_ids = []
-        field = {
-            "visibility": visibility,
-            "visibleUserIds": visible_user_ids,
-            "text": content,
-            "cw": cw,
-            "localOnly": local_only,
-            "noExtractMentions": no_extract_mentions,
-            "noExtractHashtags": no_extract_hashtags,
-            "noExtractEmojis": no_extract_emojis,
-            "replyId": reply_id,
-            "renoteId": renote_id,
-            "channelId": channel_id
-        }
-        if not check_multi_arg(content, file_ids, renote_id, poll):
-            raise ContentRequired("ノートの送信にはcontent, file_ids, renote_id またはpollのいずれか1つが無くてはいけません")
-
-        if poll and type(Poll):
-            poll_data = remove_dict_empty({
-                'choices': poll.choices,
-                'multiple': poll.multiple,
-                'expiresAt': poll.expires_at,
-                'expiredAfter': poll.expired_after
-            })
-            field["poll"] = poll_data
-
-        if file_ids:
-            field["fileIds"] = file_ids
-        field = remove_dict_empty(field)
-        res = await self.http.request(Route('POST', '/api/notes/create'), json=field, auth=True, lower=True)
-        return Note(res["created_note"], state=self)
-
-    async def create_renote(self, note_id: str) -> Note:
-        return await self.post_note(renote_id=note_id)
-
-    async def create_quote(self, note_id: str,
-                           content: Optional[str] = None,
-                           visibility: str = 'public',
-                           visible_user_ids: Optional[List[str]] = None,
-                           cw: Optional[str] = None,
-                           local_only: bool = False,
-                           no_extract_mentions: bool = False,
-                           no_extract_hashtags: bool = False,
-                           no_extract_emojis: bool = False,
-                           file_ids=None,
-                           poll: Optional[Poll] = None) -> Note:
-        return await self.post_note(content=content, visibility=visibility, visible_user_ids=visible_user_ids, cw=cw,
-                                    local_only=local_only, no_extract_mentions=no_extract_mentions,
-                                    no_extract_hashtags=no_extract_hashtags,
-                                    no_extract_emojis=no_extract_emojis, renote_id=note_id, file_ids=file_ids, poll=poll)
-
-    async def get_note(self, note_id) -> Note:
-        res = await self.http.request(Route('POST', '/api/notes/show'), json={"noteId": note_id}, auth=True, lower=True)
-        return Note(res, state=self)
-
-    async def get_replies(self, note_id: str, since_id: Optional[str] = None, until_id: Optional[str] = None, limit: int = 10,) -> List[Note]:
-        res = await self.http.request(Route('POST', '/api/notes/replies'), json={"noteId": note_id, "sinceId": since_id, "untilId": until_id, "limit": limit}, auth=True, lower=True)
-        return [Note(i, state=self) for i in res]
+    def get_note_instance(self, note_id: str) -> NoteActions:
+        return NoteActions(self.__state, self.__http, self.__loop, note_id=note_id)
 
 
-class UserAction:
-    def __init__(self, http: HTTPClient, loop: asyncio.AbstractEventLoop):
-        self.http = http
-        self.loop = loop
-
-    async def accept_following_request(self, user_id: str) -> bool:
-        data = {'userId': user_id}
-        return bool(await self.http.request(Route('POST', '/api/following/requests/accept'), json=data, auth=True))
-
-    async def reject_following_request(self, user_id: str) -> bool:
-        data = {'userId': user_id}
-        return bool(await self.http.request(Route('POST', '/api/following/requests/reject'), json=data, auth=True))
-
-
-class ClientAction(NoteActions, UserAction):
-    pass
-
-
-class ConnectionState(ClientAction):
+class ConnectionState(ClientActions):
     def __init__(self, dispatch: Callable[..., Any], http: HTTPClient, loop: asyncio.AbstractEventLoop, client: Client):
-        super().__init__(http, loop)
+        super().__init__(self, http, loop)
         self.client: Client = client
         self.dispatch = dispatch
         self.http: HTTPClient = http
@@ -195,8 +62,10 @@ class ConnectionState(ClientAction):
             if attr.startswith('parse'):
                 parsers[attr[6:].upper()] = func
 
+    def get_client_actions(self) -> ClientActions:
+        return ClientActions(self, self.http, self.loop)
+
     def parse_emoji_added(self, message: Dict[str, Any]):
-        print(message)
         self.dispatch('emoji_add', Emoji(message['body']['emoji'], state=self))
 
     def parse_channel(self, message: Dict[str, Any]) -> None:
@@ -223,12 +92,17 @@ class ConnectionState(ClientAction):
         フォローを解除した際のイベントを解析する関数
         """
 
+    def parse_signin(self, message: Dict[str, Any]):
+        """
+        ログインが発生した際のイベント
+        """
+
     def parse_receive_follow_request(self, message: Dict[str, Any]):
         """
         フォローリクエストを受け取った際のイベントを解析する関数
         """
 
-        self.dispatch('follow_request', Following(message, state=self))
+        self.dispatch('follow_request', FollowRequest(message, state=self))
 
     def parse_me_updated(self, message: Dict[str, Any]):
         pass
@@ -240,28 +114,28 @@ class ConnectionState(ClientAction):
         """
         リプライ
         """
-        self.dispatch('message', Note(message, state=self))
+        self.dispatch('message', Note(RawNote(message), state=self))
 
     def parse_follow(self, message: Dict[str, Any]) -> None:
         """
         ユーザーをフォローした際のイベントを解析する関数
         """
 
-        self.dispatch('user_follow', User(message, state=self))
+        self.dispatch('user_follow', User(RawUser(message), state=self))
 
     def parse_followed(self, message: Dict[str, Any]) -> None:
         """
         フォローイベントを解析する関数
         """
 
-        self.dispatch('follow', User(message, state=self))
+        self.dispatch('follow', User(RawUser(message), state=self))
 
     def parse_mention(self, message: Dict[str, Any]) -> None:
         """
         メンションイベントを解析する関数
         """
 
-        self.dispatch('mention', Note(message, state=self))
+        self.dispatch('mention', Note(RawNote(message), state=self))
 
     def parse_drive_file_created(self, message: Dict[str, Any]) -> None:
         pass  # TODO: 実装
@@ -281,6 +155,9 @@ class ConnectionState(ClientAction):
     def parse_unread_mention(self, message: Dict[str, Any]) -> None:
         pass
 
+    def parse_unread_specified_note(self, message: Dict[str, Any]) -> None:
+        pass
+
     def parse_read_all_messaging_messages(self, message: Dict[str, Any]) -> None:
         pass
 
@@ -288,13 +165,13 @@ class ConnectionState(ClientAction):
         """
         チャットが来た際のデータを処理する関数
         """
-        self.dispatch('message', Chat(message, state=self))
+        self.dispatch('message', Chat(RawChat(message), state=self))
 
     def parse_unread_messaging_message(self, message: Dict[str, Any]) -> None:
         """
         チャットが既読になっていない場合のデータを処理する関数
         """
-        self.dispatch('message', Chat(message, state=self))
+        self.dispatch('message', Chat(RawChat(message), state=self))
 
     def parse_notification(self, message: Dict[str, Any]) -> None:
         """
@@ -304,11 +181,8 @@ class ConnectionState(ClientAction):
         ----------
         message: Dict[str, Any]
             Received message
-
-        Returns
-        -------
-        None
         """
+
         accept_type = ['reaction']
         notification_type = str_lower(message['type'])
         if notification_type in accept_type:
@@ -338,57 +212,17 @@ class ConnectionState(ClientAction):
         """
         self.dispatch('reaction', Reaction(message, state=self))
 
-    def parse_note(self, message: Dict[str, Any]) -> None:
+    def parse_note(self, message: NotePayload) -> None:
         """
         ノートイベントを解析する関数
         """
-        note = Note(message, state=self)
+        note = Note(RawNote(message), state=self)
         # Router(self.http.ws).capture_message(note.id) TODO: capture message
         self.client._on_message(note)
 
-    async def follow_user(self, user_id: str) -> tuple[bool, Optional[str]]:
-        """
-        与えられたIDのユーザーをフォローします
-
-        Parameters
-        ----------
-        user_id : Optional[str] = None
-            フォローしたいユーザーのID
-
-        Returns
-        -------
-        status: bool = False
-            成功ならTrue, 失敗ならFalse
-        """
-        data = {"userId": user_id}
-        res = await self.http.request(Route('POST', '/api/following/create'), json=data, auth=True, lower=True)
-        if res.get("error"):
-            code = res["error"]["code"]
-            status = False
-        else:
-            code = None
-            status = True
-        return status, code
-
-    async def unfollow_user(self, user_id: str) -> bool:
-        """
-        Parameters
-        ----------
-        user_id :
-            フォローを解除したいユーザーのID
-
-        Returns
-        -------
-        status: bool = False
-            成功したならTrue, 失敗したならFalse
-        """
-        data = {"userId": user_id}
-        res = await self.http.request(Route('POST', '/api/following/delete'), json=data, auth=True)
-        return bool(res.status_code == 204 or 200)
-
     async def get_i(self) -> User:
         res = await self.http.request(Route('POST', '/api/i'), auth=True, lower=True)
-        return User(res, state=self)
+        return User(RawUser(res), state=self)
 
     def get_users(self,
                   limit: int = 10,
@@ -403,43 +237,6 @@ class ConnectionState(ClientAction):
         return InstanceIterator(self).get_users(limit=limit, offset=offset, sort=sort, state=state, origin=origin,
                                                 username=username, hostname=hostname, get_all=get_all)
 
-    async def add_reaction(self, reaction: str, note_id: Optional[str] = None) -> bool:
-        """
-        指定したnoteに指定したリアクションを付与します（内部用
-
-        Parameters
-        ----------
-        reaction : Optional[str]
-            付与するリアクション名
-        note_id : Optional[str]
-            付与対象のノートID
-
-        Returns
-        -------
-        status: bool
-            成功したならTrue,失敗ならFalse
-        """
-        data = remove_dict_empty({"noteId": note_id, "reaction": reaction})
-        return await self.http.request(Route('POST', '/api/notes/reactions/create'), json=data, auth=True, lower=True)
-
-    async def delete_note(self, note_id: str) -> bool:
-        """
-        指定したidのノートを削除します。
-
-        Parameters
-        ----------
-        note_id : str
-            ノートid
-
-        Returns
-        -------
-        bool
-            成功したか否か
-        """
-        data = {"noteId": note_id}
-        res = await self.http.request(Route('POST', '/api/notes/delete'), json=data, auth=True)
-        return bool(res)
-
     @cached(ttl=10, namespace='get_user', key_builder=key_builder)
     async def get_user(self, user_id: Optional[str] = None, username: Optional[str] = None,
                        host: Optional[str] = None) -> User:
@@ -458,12 +255,13 @@ class ConnectionState(ClientAction):
 
         Returns
         -------
-        dict:
+        dict
             ユーザー情報
         """
+
         field = remove_dict_empty({"userId": user_id, "username": username, "host": host})
         data = await self.http.request(Route('POST', '/api/users/show'), json=field, auth=True, lower=True)
-        return User(data, state=self)
+        return User(RawUser(data), state=self)
 
     async def post_chat(self, content: str, *, user_id: str = None, group_id: str = None, file_id=None) -> Chat:
         """
@@ -487,7 +285,7 @@ class ConnectionState(ClientAction):
         """
         args = remove_dict_empty({'userId': user_id, 'groupId': group_id, 'text': content, 'fileId': file_id})
         data = await self.http.request(Route('POST', '/api/messaging/messages/create'), json=args, auth=True, lower=True)
-        return Chat(data, state=self)
+        return Chat(RawChat(data), state=self)
 
     async def delete_chat(self, message_id: str) -> bool:
         """
@@ -524,7 +322,7 @@ class ConnectionState(ClientAction):
 
         Returns
         -------
-        dict:
+        dict
             ユーザー情報
         """
         if not check_multi_arg(user_id, username):
@@ -534,7 +332,7 @@ class ConnectionState(ClientAction):
         data = await self.http.request(Route('POST', '/api/users/show'), json=field, auth=True, lower=True)
         old_cache = Cache(namespace='get_user')
         await old_cache.delete(kwargs['cache_key'].format('get_user'))
-        return User(data, state=self)
+        return User(RawUser(data), state=self)
 
     async def get_followers(
             self,
@@ -545,7 +343,7 @@ class ConnectionState(ClientAction):
             until_id: Optional[str] = None,
             limit: int = 10,
             get_all: bool = False,
-    ) -> AsyncIterator[Follower]:
+    ) -> AsyncIterator[Followee]:
         """
         与えられたユーザーのフォロワーを取得します
 
@@ -597,11 +395,11 @@ class ConnectionState(ClientAction):
                     data["untilId"] = get_data[-1]["id"]
                 else:
                     break
-                for i in [Follower(i, state=self) for i in get_data]:
+                for i in [Followee(i, state=self) for i in get_data]:
                     yield i
         else:
             get_data = await self.http.request(Route('POST', '/api/users/followers'), json=data, auth=True, lower=True)
-            for i in [Follower(i, state=self) for i in get_data]:
+            for i in [Followee(i, state=self) for i in get_data]:
                 yield i
 
     @cached(ttl=10, key_builder=key_builder, key='get_instance')
@@ -622,14 +420,6 @@ class ConnectionState(ClientAction):
     async def remove_emoji(self, emoji_id: str) -> bool:
         return bool(await self.http.request(Route('POST', '/api/admin/emoji/remove'), json={'id': emoji_id}, auth=True))
 
-    async def show_file(self, file_id: Optional[str], url: Optional[str]) -> Drive:
-        data = remove_dict_empty({"fileId": file_id, "url": url})
-        return Drive(await self.http.request(Route('POST', '/api/admin/drive/show-file'), json=data, auth=True, lower=True),
-                     state=self)
-
-    async def remove_file(self, file_id: str) -> bool:
-        return bool(await self.http.request(Route('POST', '/api/drive/files/delete'), json={'fileId': file_id}, auth=True))
-
     async def get_user_notes(
             self,
             user_id: str,
@@ -645,7 +435,7 @@ class ConnectionState(ClientAction):
             file_type: Optional[List[str]] = None,
             since_date: int = 0,
             until_data: int = 0
-    ) -> AsyncIterator[Note]:
+    ) -> Generator[Note]:
         if limit > 100:
             raise InvalidParameters("limit は100以上を受け付けません")
 
@@ -672,11 +462,11 @@ class ConnectionState(ClientAction):
                     break
                 args["untilId"] = get_data[-1]["id"]
                 for data in get_data:
-                    yield Note(NotePayload(**data), state=self)
+                    yield Note(RawNote(data), state=self)
         else:
             get_data = await self.http.request(Route('POST', '/api/users/notes'), json=args, auth=True, lower=True)
             for data in get_data:
-                yield Note(NotePayload(**upper_to_lower(data)), state=self)
+                yield Note(RawNote(**upper_to_lower(data)), state=self)
 
     async def get_announcements(self, limit: int, with_unreads: bool, since_id: str, until_id: str):
         """
@@ -689,10 +479,6 @@ class ConnectionState(ClientAction):
         since_id: str
         until_id: str
             前回の最後の値を与える(既に実行し取得しきれない場合に使用)
-
-        Returns
-        -------
-
         """
 
         if limit > 100:
@@ -714,7 +500,7 @@ class ConnectionState(ClientAction):
             *,
             force: bool = False,
             is_sensitive: bool = False,
-    ) -> Drive:
+    ) -> File:
 
         if to_file and to_url is None:  # ローカルからアップロードする
             with open(to_file, "rb") as f:
@@ -725,4 +511,4 @@ class ConnectionState(ClientAction):
             res = await self.http.request(Route('POST', '/api/drive/files/upload-from-url'), json=args, auth=True, lower=True)
         else:
             raise InvalidParameters("path または url のどちらかは必須です")
-        return Drive(res, state=self)
+        return File(RawFile(res), state=self)
